/*
 * Copyright (C) 2000 Fujio Nobori (toh@fuji-climb.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: PortForwarder.h,v 1.14 2002/02/06 23:28:44 toh Exp $
 */

#ifndef PORTFORWARDER_H
#define PORTFORWARDER_H

#include <windows.h>
#include <winsock.h>
#include <tchar.h>

/*****************************************************************/
/* for OpenSSL and OpenSSH                                       */
/*****************************************************************/
#define close(s) closesocket(s)

#ifndef _WIN32_WCE
typedef long time_t;        /* time value */
#endif /* _WIN32_WCE */

int write(SOCKET s,
          void *buf,
          int len);
int read(SOCKET s,
         void *buf,
         int len);

/* Copies LPSTR string to LPTSTR buffer. */
void lptstrCopy(LPTSTR buf,
                int length,
                LPCSTR string);
/* Copies LPTSTR string to LPSTR buffer. */
void lpstrCopy(LPSTR buf,
               int length,
               LPCTSTR string);
/* Creates LPTSTR string from LPSTR string. */
LPTSTR lptstrCreate(LPCSTR string);
/* Creates LPSTR string from LPTSTR string. */
LPSTR lpstrCreate(LPCTSTR string);

#ifdef _WIN32_WCE 
#if _WIN32_WCE < 300
int isupper(int c);
#endif /* _WIN32_WCE < 300 */
#else /* _WIN32_WCE */
int sscanf(const char *str,
           const char *format,
           ...);
int sprintf(char *str,
            const char *format,
            ...);
#endif /* _WIN32_WCE */

/*****************************************************************/
/* for OpenSSL                                                   */
/*****************************************************************/
#ifdef _OPENSSL_
#define BUFSIZ  512
#define assert(e) ((void)0)
#define WSACancelBlockingCall() ((void)0)

typedef _int64 int64_t;
typedef int64_t quad_t;
typedef quad_t off_t;

time_t time(time_t *tloc);

off_t lseek(int fildes,
            off_t offset,
            int whence);

void *signal(int sig, void (*func)());

#ifdef _WIN32_WCE
#if _WIN32_WCE < 300
int isdigit(int c);
int isalnum(int c);
int isspace(int c);
#endif /* _WIN32_WCE < 300 */

struct tm *gmtime(const time_t *clock);
struct tm *localtime(const time_t *clock);

struct servent *getservbyname(const char *name,
                              const char *proto);

void abort(void);

char *getenv(const char *name);

unsigned long int strtoul(const char *nptr,
                          char **endptr,
                          int base);
#endif /* _WIN32_WCE */
#endif /* _OPENSSL_ */

/*****************************************************************/
/* for OpenSSH                                                   */
/*****************************************************************/
#ifdef _OPENSSH_
#include "FmtMsg.h"

#define APP_NAME TEXT("PortForwarder")
#define APP_NAME2 TEXT("PF-keygen")
#define MAX_HOST_NUM 10
#define PF_TASKBAR_MSG (WM_USER + 100)
#define PF_CONNECTED_MSG (WM_USER + 101)
#define PF_FATALCALLED_MSG (WM_USER + 102)
#define PF_GETSTRING_MSG (WM_USER + 103)
#define PF_SHOWTTYMSGDLG_MSG (WM_USER + 104)
#define PF_SHOWMESSAGEBOX_MSG (WM_USER + 105)
#define PF_ADDTTYMESSAGE_MSG (WM_USER + 106)

#define PF_STATUS_NOT_CONNECTED 0
#define PF_STATUS_CONNECTING 1
#define PF_STATUS_CONNECTED 2
#define PF_STRING_NOT_CONNECTED TEXT("Not connected")
#define PF_STRING_CONNECTING TEXT("Connecting...")
#define PF_STRING_CONNECTED TEXT("Connected")
#define PF_STRING_DISCONNECTING TEXT("Disconnecting...")

#define NI_MAXHOST 1025
#define NI_MAXSERV 32
#define MAXHOSTNAMELEN 256
#define MAX_INPUT 1024
#define MAX_LINE 1024
#define IN_LOOPBACKNET 127
#define SHUT_RD 0
#define SHUT_WR 1
#define SHUT_RDWR 2
#define EAGAIN WSAEWOULDBLOCK /* I'm not sure if this is a correct choice... */
#define EINTR WSAEINTR

#define inline _inline
#define log LOG /* log() conflicts with one in math.h */
#define strerror(errno) GetErrMsg()
#define errno WSAGetLastError()
#define rand rnd

typedef unsigned char u_int8_t;
typedef unsigned short u_int16_t;
typedef unsigned int u_int32_t;
typedef u_int32_t uid_t;
typedef int ssize_t;
typedef unsigned int socklen_t;
typedef unsigned short u_short;

/* Struct for input dialog. */
struct inputinfo {
    LPTSTR string;
    BOOL passwd;
    char inputstring[MAX_INPUT];
};

/* Struct for MessageBox info. */
struct msgboxinfo {
    LPTSTR caption;
    LPTSTR message;
    UINT type;
    UINT result;
};

/* Gets some values from ini file. */
BOOL LoadInfo();

/* Sets some values to ini file. */
void SaveInfo();

#ifdef _DEBUG
/* Show messages. Debugging purpose. */
void TestMessage(const char *keyword,
                 const char *string);
#endif /* _DEBUG */

/* Shows message. */
void ShowMessage(LPCTSTR fmt,...);

/* Shows input dialog and returns the input. */
char *GetInputString(char *buf,
                     unsigned int len,
                     LPTSTR caption,
                     BOOL passwd);

/* Shows input dialog and returns the input. */
char *GetInputString2(char *buf,
                      unsigned int len,
                      LPTSTR message,
                      BOOL passwd);

/* Forces to exit the program. */
void Exit();

/* Substitute for fgets(). */
char* GetLine(char *buf,
              int size,
              HANDLE file);

void exit(int status);

#ifndef _WIN32_WCE
int vsprintf(char *buffer,
             const char *fmt,
             va_list argptr);
#endif /* _WIN32_WCE */

#ifdef _WIN32_WCE
size_t strspn(const char *string1,
              const char *string2);
#endif /* _WIN32_WCE */

int strcasecmp(const char *s1, const char *s2);

long string_to_long(const char* nptr,
                    char **endptr,
                    int base);

#ifdef _LOG_
/* Opens log file. */
void OpenLogFile();

/* Closes log file. */
void CloseLogFile();

/* Logs message. */
void MyLog(const char *keyword,
            LPCTSTR message);
#endif /* _LOG_ */

/* Gets error message from system. Seems not to work on WinCE?? */
LPTSTR GetErrMsg();

/* Main function of OpenSSH. */
void main();

/* Sends SSH_CMSG_EOF to server. */
void close_now();

/* Sends string to server. */
void send_string_to_server(const char *string);

/* Shows dialog box when stdout or stderr message comes. */
void handle_tty_message(const char *message);

/* Shows open-file dialog and gets file name. */
BOOL FindFile(HWND owner,
              LPTSTR title,
              LPTSTR filter,
              LPTSTR dir,
              LPTSTR file,
              int length);

int remove(const char *path);

unsigned int sleep(unsigned int seconds);
#endif /* _OPENSSH_ */

#endif /* PORTFORWARDER_H */
