/*
 * Copyright (C) 2000 Fujio Nobori (toh@fuji-climb.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: TohCommon.c,v 1.13 2002/02/06 23:28:45 toh Exp $
 */

#include <windows.h>
#include <winsock.h>
#include <tchar.h>
#include "PortForwarder.h"
#include "resource.h"
#include "ssh.h"

static TCHAR errMsg[1024];

extern HANDLE g_hWnd;
extern HINSTANCE g_hInst;

#ifdef _LOG_
static HANDLE logFile;
static BOOL logFlag = FALSE;
#endif /* _LOG_ */

/* Exit program when called. */
void abort(void)
{
    fatal("aborted");
}

/* Copies LPSTR string to LPTSTR buffer. */
void lptstrCopy(LPTSTR buf,
                int length,
                LPCSTR string)
{
#ifdef _WIN32_WCE
    int len;

    len = MultiByteToWideChar(CP_ACP, 0, string, -1, buf, length);
    buf[len] = TEXT('0');
#else /* _WIN32_WCE */
    strcpy(buf, string);
#endif /* _WIN32_WCE */
}

/* Copies LPTSTR string to LPSTR buffer. */
void lpstrCopy(LPSTR buf,
               int length,
               LPCTSTR string)
{
#ifdef _WIN32_WCE
    int len;

    len = WideCharToMultiByte(CP_ACP, 0, string, -1, buf, length, NULL, NULL);
    buf[len] = '\0';
#else /* _WIN32_WCE */
    strcpy(buf, string);
#endif /* _WIN32_WCE */
}

/* Creates LPTSTR string from LPSTR string. */
LPTSTR lptstrCreate(LPCSTR string)
{
    int len;
    LPTSTR temp;

#ifdef _WIN32_WCE
    len = MultiByteToWideChar(CP_ACP, 0, string, -1, NULL, 0);
    temp = (LPTSTR)malloc((len + 1) * sizeof(TCHAR));
    MultiByteToWideChar(CP_ACP, 0, string, -1, temp, len);
    temp[len] = TEXT('0');
#else /* _WIN32_WCE */
    len = strlen(string);
    temp = (LPTSTR)malloc((len + 1) * sizeof(TCHAR));
    strcpy(temp, string);
#endif /* _WIN32_WCE */

    return temp;
}

/* Creates LPSTR string from LPTSTR string. */
LPSTR lpstrCreate(LPCTSTR string)
{
    int len;
    LPSTR temp;

#ifdef _WIN32_WCE
    len = WideCharToMultiByte(CP_ACP, 0, string, -1, NULL, 0, NULL, NULL);
    temp = (LPSTR)malloc(len + 1);
    WideCharToMultiByte(CP_ACP, 0, string, -1, temp, len, NULL, NULL);
    temp[len] = '\0';
#else /* _WIN32_WCE */
    len = _tcslen(string);
    temp = (LPSTR)malloc(len + 1);
    strcpy(temp, string);
#endif /* _WIN32_WCE */

    return temp;
}

#ifdef _LOG_
/* Opens log file. */
void OpenLogFile()
{
    TCHAR filename[] = TEXT("\\Temp\\PortForwarder.log");
    char startMessage[] = "#### PortForwarder started\n";
    char dateMessage[]  = "####   DATE ";
    char timeMessage[]  = "####   TIME ";
    TCHAR string[256];
    char string2[256];
    DWORD sizeWritten;

    if (logFlag)
        return;
    logFile = CreateFile(filename, GENERIC_WRITE, FILE_SHARE_READ, NULL,
                         OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (logFile == INVALID_HANDLE_VALUE && g_hWnd) {
        logFlag = FALSE;
        MessageBox(g_hWnd, TEXT("log is unavailable."),
                   APP_NAME, MB_OK | MB_ICONINFORMATION);
        return;
    }

    SetFilePointer(logFile, 0, NULL, FILE_END);

    WriteFile(logFile, startMessage, strlen(startMessage), &sizeWritten, NULL);
    WriteFile(logFile, dateMessage, strlen(dateMessage), &sizeWritten, NULL);
    GetDateFormat(LOCALE_SYSTEM_DEFAULT, 0, NULL, NULL, string, 256);
    lpstrCopy(string2, sizeof(string2), string);
    WriteFile(logFile, string2, strlen(string2), &sizeWritten, NULL);
    WriteFile(logFile, "\n", strlen("\n"), &sizeWritten, NULL);
    WriteFile(logFile, timeMessage, strlen(timeMessage), &sizeWritten, NULL);
    GetTimeFormat(LOCALE_SYSTEM_DEFAULT, 0, NULL, NULL, string, 256);
    lpstrCopy(string2, sizeof(string2), string);
    WriteFile(logFile, string2, strlen(string2), &sizeWritten, NULL);
    WriteFile(logFile, "\n", strlen("\n"), &sizeWritten, NULL);

    logFlag = TRUE;
}

/* Closes log file. */
void CloseLogFile()
{
    char dateMessage[]  = "####   DATE ";
    char timeMessage[]  = "####   TIME ";
    TCHAR string[256];
    char string2[256];
    DWORD sizeWritten;

    if (!logFlag)
        return;

    WriteFile(logFile, dateMessage, strlen(dateMessage), &sizeWritten, NULL);
    GetDateFormat(LOCALE_SYSTEM_DEFAULT, 0, NULL, NULL, string, 256);
    lpstrCopy(string2, sizeof(string2), string);
    WriteFile(logFile, string2, strlen(string2), &sizeWritten, NULL);
    WriteFile(logFile, "\n", strlen("\n"), &sizeWritten, NULL);
    WriteFile(logFile, timeMessage, strlen(timeMessage), &sizeWritten, NULL);
    GetTimeFormat(LOCALE_SYSTEM_DEFAULT, 0, NULL, NULL, string, 256);
    lpstrCopy(string2, sizeof(string2), string);
    WriteFile(logFile, string2, strlen(string2), &sizeWritten, NULL);
    WriteFile(logFile, "\n", strlen("\n"), &sizeWritten, NULL);
    WriteFile(logFile, "\n\n", strlen("\n\n"), &sizeWritten, NULL);
    CloseHandle(logFile);
    logFlag = FALSE;
}

/* Logs message. */
void MyLog(const char *keyword,
            LPCTSTR message)
{
    DWORD sizeWritten;
    char *message2;

    if (!logFlag)
        return;

    message2 = lpstrCreate(message);

    WriteFile(logFile, keyword, strlen(keyword), &sizeWritten, NULL);
    WriteFile(logFile, message2, strlen(message2), &sizeWritten, NULL);
    WriteFile(logFile, "\n", strlen("\n"), &sizeWritten, NULL);

    FlushFileBuffers(logFile);

    free(message2);
}
#endif /* _LOG_ */

#ifdef _DEBUG
/* Show messages. Debugging purpose. */
void TestMessage(const char *keyword,
                 const char *string)
{
    LPTSTR keyword2;
    LPTSTR string2;
    LPTSTR temp;

    keyword2 = lptstrCreate(keyword);
    string2 = lptstrCreate(string);
    temp = FmtMsg(TEXT("%1!s! = <%2!s!>"), keyword2, string2);
    SetForegroundWindow(g_hWnd);
    MessageBox(g_hWnd, temp, TEXT("TEST"), MB_OK);
    free(keyword2);
    free(string2);
    LocalFree(temp);
}
#endif /* _DEBUG */

/* Shows message. */
void ShowMessage(LPCTSTR fmt,...)
{
    TCHAR buf[1024];
    struct msgboxinfo info;
    va_list ap;

    va_start(ap, fmt);
    _vstprintf(buf, fmt, ap);
    va_end(ap);

    info.caption = APP_NAME TEXT(":Message");
    info.message = buf;
    info.type = MB_OK | MB_ICONINFORMATION;
    SendMessage(g_hWnd, PF_SHOWMESSAGEBOX_MSG, (WPARAM)&info, 0);
}

/* Gets error message from system. Seems not to work on WinCE?? */
LPTSTR GetErrMsg()
{
    DWORD rtn;
    LPVOID lpMsgBuf;
    rtn = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
                        FORMAT_MESSAGE_FROM_SYSTEM | 
                        FORMAT_MESSAGE_IGNORE_INSERTS,
                        NULL,
                        GetLastError(),
                        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                        (LPTSTR) &lpMsgBuf,
                        0,
                        NULL);

    if (rtn) {
        _tcscpy(errMsg, (LPTSTR)lpMsgBuf);
        LocalFree(lpMsgBuf);
    }
    else {
        _tcscpy(errMsg, TEXT("An error has been encountered."));
    }

    return errMsg;
}

/* Shows input dialog and returns the input. */
char *GetInputString(char *buf,
                     unsigned int len,
                     LPTSTR caption,
                     BOOL passwd)
{
    struct inputinfo info;

    info.string = caption;
    info.passwd = passwd;

    SendMessage(g_hWnd, PF_GETSTRING_MSG, (WPARAM)&info, 0);

    if (strlen(info.inputstring) > (len -1)) {
        fatal("String is longer than expected.");
        return NULL;
    }

    strcpy(buf, info.inputstring);
    return buf;
}

/* Returns time. */
time_t time(time_t *tloc)
{
    SYSTEMTIME systemTime;
    time_t temp;

    GetSystemTime(&systemTime);
    temp = (systemTime.wYear - 1970) * 365 * 24 * 60 * 60
        + systemTime.wMonth * 30 * 24 * 60 * 60
        + systemTime.wDay * 24 * 60 * 60
        + systemTime.wHour * 60 * 60
        + systemTime.wMinute * 60
        + systemTime.wSecond;
    if (tloc)
        *tloc = temp;

    return temp;
}

/* Calles send(). */
int write(SOCKET s,
          void *buf,
          int len)
{
    return send(s, buf, len, 0);
}

/* Calles recv(). */
int read(SOCKET s,
         void *buf,
         int len)
{
    return recv(s, buf, len, 0);
}

/* Converts string value to long. */
long string_to_long(const char* nptr,
                    char **endptr,
                    int base)
{
    LPTSTR nptr2;
    LPTSTR endptr2;
    long value;

    nptr2 = lptstrCreate(nptr);

    value = _tcstol(nptr2, &endptr2, base);
    *endptr = (char *)nptr + (endptr2 - nptr2);
    free(nptr2);
    return value;
}

/* Compares strings, ignoring case. */
int strcasecmp(const char *s1, const char *s2)
{
    char *str1, *str2;
    int i;

    str1 = (char *)malloc(strlen(s1) + 1);
    strcpy(str1, s1);
    str2 = (char *)malloc(strlen(s2) + 1);
    strcpy(str2, s2);

    for (i = 0; str1[i] != '\0'; i ++) {
        if (str1[i] >= 'A' && str1[i] <= 'Z')
            str1[i] += ('a' - 'A');
    }

    for (i = 0; str2[i] != '\0'; i ++) {
        if (str2[i] >= 'A' && str2[i] <= 'Z')
            str2[i] += ('a' - 'A');
    }

    i = strcmp(str1, str2);

    free(str1);
    free(str2);

    return i;
}

#ifdef _WIN32_WCE
#if _WIN32_WCE < 300
/* Returns true if decimal-digit character is passed. */
int isdigit(int c)
{
    if ('0' <= c && c <= '9')
        return TRUE;
    else
        return FALSE;
}

/* Returns true if upper-case character is passed. */
int isupper(int c)
{
    if ('A' <= c && c <= 'Z')
        return TRUE;
    else
        return FALSE;
}
#endif /* _WIN32_WCE < 300 */

/* Spans a string. */
size_t strspn(const char *string1,
              const char *string2)
{
    size_t i, j;

    for (i = 0; string1[i] != '\0'; i ++) {
        for (j = 0; string2[j] != '\0'; j ++)
            if (string2[j] == string1[i]) break;
        if (string2[j] == '\0')
            return i;
    }

    return i;
}
#endif /* _WIN32_WCE */

/* Sleeps. Zzz... */
unsigned int sleep(unsigned int seconds)
{
    Sleep(seconds * 1000);
    return 0;
}

/* Removes path. */
int remove(const char *path)
{
    LPTSTR path2;
    int value;

    path2 = lptstrCreate(path);
    value = (int)DeleteFile(path2);

    free(path2);
    return value;
}
