/*
 * Copyright (C) 2000 Fujio Nobori (toh@fuji-climb.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pfmain.c,v 1.18 2002/02/06 23:28:45 toh Exp $
 */

#include <windows.h>
#include "PortForwarder.h"
#include "resource.h"
#include "ssh.h"
#include "packet.h"
#if (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300))
#include <aygshell.h>
#endif /* (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300)) */

/* Handle to current instance. */
HINSTANCE g_hInst;
/* Handle to main dialog. */
HWND g_hWnd;
/* Handle to tty message dialog. */
HWND g_hTtyMsgDlg = NULL;
/* Handle to the main thread. */
static HANDLE g_hMainThread = NULL;
/* Handle to the client_loop thread. */
static HANDLE g_hClientLoopThread = NULL;
/* Flag indicating whether user requests to show tty messages. */
BOOL g_showMessage;
/* Hosts that were most currently used. */
TCHAR g_hosts[MAX_HOST_NUM][MAX_PATH];
/* Statuses of show message that were most currently used. */
BOOL g_showMessages[MAX_HOST_NUM];
/* Path of config folder. */
TCHAR g_configFolder[MAX_PATH];
/* Name of knowon-hosts file. */
TCHAR g_knownHostsFile[MAX_PATH];
/* Name of identity file. */
TCHAR g_identityFile[MAX_PATH];
/* Name of config file. */
TCHAR g_configFile[MAX_PATH];
/* Flag indicating whether user requests to hide main dialog automatically. */
BOOL g_automaticallyHide;

/* Name of the connecting host. */
static TCHAR hostTemp[NI_MAXHOST];
/* Flag indicating whether WSAStartup() was called successfully. */
static int wsaStartup = 1;
/* Flag indicating whether the icon is in the taskbar status area. */
static BOOL iconFlag = FALSE;
/* Status of connection. */
static int connectionStatus = PF_STATUS_NOT_CONNECTED;

/* Name of the host we are connecting to. */
extern char *host;

/* Flag indicating whether user requests to close the connection. */
extern BOOL closeNowFlag;

/* Critical section for Popup menu. */
static CRITICAL_SECTION critical_section;

/* Get option value */
LPTSTR GetOption(LPTSTR line,
                 LPTSTR string);

/* Calls client_loop(). This function is to be called by CreateThread(). */
static DWORD WINAPI DoClientLoop(PVOID pArg)
{
    client_loop(0, 0);
#if !(defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300))
    /* Somehow my PocketPC fails to do close(connection_out)
     * in packet_close(). Why?
     */
    packet_close();
#endif /* !(defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300)) */

    if (!closeNowFlag) {
        struct msgboxinfo info;
        info.caption = APP_NAME TEXT(":FATAL ERROR");
        info.message = TEXT("Connection closed by server.");
        info.type = MB_OK | MB_ICONERROR;
        SendMessage(g_hWnd, PF_SHOWMESSAGEBOX_MSG, (WPARAM)&info, 0);
        SendMessage(g_hWnd, PF_FATALCALLED_MSG, 0, 0);
    }

    ExitThread(0);

    return 0;
}

/* Calls main(). This function is to be called by Connect(). */
static DWORD WINAPI DoMain(PVOID pArg)
{
    connectionStatus = PF_STATUS_CONNECTING;
    main();
    connectionStatus = PF_STATUS_CONNECTED;

    SendMessage(g_hWnd, PF_CONNECTED_MSG, 0, 0);

    ExitThread(0);

    return 0;
}

/* Connect to the host. */
static void Connect()
{
    DWORD threadID;

    SendDlgItemMessage(g_hWnd, IDC_PF_STATUSSTATIC, WM_SETTEXT,
                       (WPARAM)0, (LPARAM)PF_STRING_CONNECTING);

    g_hMainThread = CreateThread(NULL, 0, DoMain, NULL, 0, &threadID);
}

/* Adjusts file names. */
static void AdjustFileNames()
{
    TCHAR buf[MAX_PATH];
    BOOL rootFlag;

    /* Check if config folder is root directory. */
    if (g_configFolder[_tcslen(g_configFolder) - 1] == TEXT('\\')) {
        rootFlag = TRUE;
    }
    else {
        rootFlag = FALSE;
    }

    _tcscpy(buf, g_configFile);
    _tcscpy(g_configFile, g_configFolder);
    if (!rootFlag) {
        _tcscat(g_configFile, TEXT("\\"));
    }
    _tcscat(g_configFile, buf);

    _tcscpy(buf, g_identityFile);
    _tcscpy(g_identityFile, g_configFolder);
    if (!rootFlag) {
        _tcscat(g_identityFile, TEXT("\\"));
    }
    _tcscat(g_identityFile, buf);

    _tcscpy(buf, g_knownHostsFile);
    _tcscpy(g_knownHostsFile, g_configFolder);
    if (!rootFlag) {
        _tcscat(g_knownHostsFile, TEXT("\\"));
    }
    _tcscat(g_knownHostsFile, buf);
}

/* Puts the icon int the taskbar status area. */
static void SetIcon()
{
    NOTIFYICONDATA nid;

    nid.cbSize = sizeof(NOTIFYICONDATA);
    nid.hWnd = g_hWnd;
    nid.uID = IDI_PF_ICON;
    nid.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
    nid.uCallbackMessage = PF_TASKBAR_MSG;
    nid.hIcon = LoadImage(g_hInst, MAKEINTRESOURCE(IDI_PF_ICON),
                          IMAGE_ICON, 16, 16, 0);
    _tcscpy(nid.szTip, hostTemp);

    Shell_NotifyIcon(NIM_ADD, &nid);
    iconFlag = TRUE;
}

/* Removes the icon from the taskbar status area. */
static void RemoveIcon()
{
    NOTIFYICONDATA nid;

    if (!iconFlag)
        return;

    memset(&nid, 0, sizeof nid);
    nid.cbSize = sizeof(NOTIFYICONDATA);
    nid.hWnd = g_hWnd;
    nid.uID = IDI_PF_ICON;

    Shell_NotifyIcon(NIM_DELETE, &nid);

    iconFlag = 0;
}

/* Wait for the client_loop thread to end. */
void PrepareExit()
{
    HANDLE threadHandles[1];
    DWORD threadCount = 1;
    DWORD rtn;

    threadHandles[0] = g_hClientLoopThread;

    while(1) {
        /* WaitForSingleObject() is not enough here, because */
        /* some window messages will be sent.                */
        rtn = MsgWaitForMultipleObjects(threadCount, threadHandles,
                                        FALSE, 1000, QS_ALLINPUT);
        if (rtn >= WAIT_OBJECT_0 && rtn < WAIT_OBJECT_0 + threadCount) {
            CloseHandle(g_hClientLoopThread);
            break;
        }
        if (rtn == WAIT_TIMEOUT) {
            TerminateThread(g_hClientLoopThread, 0);
            CloseHandle(g_hClientLoopThread);
            break;
        }
        else {
            MSG msg;
            while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
                if ((g_hWnd == NULL || !IsDialogMessage(g_hWnd, &msg)) &&
                    (g_hTtyMsgDlg == NULL ||
                     !IsDialogMessage(g_hTtyMsgDlg, &msg))) {
                    TranslateMessage(&msg);
                    DispatchMessage(&msg);
                }
            }
        }
    }
}

/* Force application to exit. */
void Exit()
{
#ifdef _LOG_
    CloseLogFile();
#endif /* _LOG_ */
    RemoveIcon();
#ifdef _WIN32_WCE
    ExitThread(0);
#else /* _WIN32_WCE */
    ExitProcess(0);
#endif /* _WIN32_WCE */
}

/* Gets information on main dialog. */
static BOOL GetInfo()
{
    SendDlgItemMessage(g_hWnd, IDC_PF_HOSTCOMBO, WM_GETTEXT,
                       (WPARAM)NI_MAXHOST, (LPARAM)hostTemp);
    if (BST_CHECKED == SendDlgItemMessage(g_hWnd, IDC_PF_SHOWMESSAGECHK,
                                          BM_GETCHECK, 1, 0)) {
        g_showMessage = TRUE;
    }
    else {
        g_showMessage = FALSE;
    }

    if (BST_CHECKED == SendDlgItemMessage(g_hWnd, IDC_PF_AUTOHIDECHK,
                                          BM_GETCHECK, 0, 0)) {
        g_automaticallyHide = TRUE;
    }
    else {
        g_automaticallyHide = FALSE;
    }

    if (_tcschr(hostTemp, TEXT(' '))) {
        SetForegroundWindow(g_hWnd);
        MessageBox(g_hWnd, TEXT("\"Host\" can't contain spaces!"), APP_NAME,
                   MB_OK | MB_ICONEXCLAMATION);
        return FALSE;
    }

    if (!_tcscmp(hostTemp, TEXT("")) ||
        !_tcscmp(g_configFolder, TEXT(""))) {
        SetForegroundWindow(g_hWnd);
        MessageBox(g_hWnd, TEXT("Fill all the blanks!"), APP_NAME,
                   MB_OK | MB_ICONEXCLAMATION);
        return FALSE;
    }

    host = lpstrCreate(hostTemp);

    return TRUE;
}

/* Insert hosts into host combo box. */
static void CreateHostsList(LPCTSTR configFile)
{
    int i;
    FILE *file;
    TCHAR line[MAX_LINE];

    SendDlgItemMessage(g_hWnd, IDC_PF_HOSTCOMBO, CB_RESETCONTENT, 0, 0);

    SendDlgItemMessage(g_hWnd, IDC_PF_HOSTCOMBO, WM_SETTEXT,
                       (WPARAM)0, (LPARAM)g_hosts[0]);

    for (i = 0; i < MAX_HOST_NUM; i ++) {
        if (!_tcscmp(g_hosts[i], TEXT(""))) {
            break;
        }
        SendDlgItemMessage(g_hWnd, IDC_PF_HOSTCOMBO, CB_ADDSTRING,
                           (WPARAM)0, (LPARAM)g_hosts[i]);
    }

    file = _tfopen(configFile, TEXT("r"));
    if (!file) {
        return;
    }

    while (_fgetts(line, MAX_LINE, file)) {
        LPTSTR cp;
        LPTSTR temp;

        temp = _tcsrchr(line, TEXT('\n'));
        if (temp) {
            *temp = TEXT('\0');
        }

        cp = line + _tcsspn(line, TEXT(" \t\r\n"));
        if (!*cp || *cp == TEXT('\n') || *cp == TEXT('#')) {
            continue;
        }

        if (_tcsncmp(cp, TEXT("HostName"), _tcslen(TEXT("HostName"))) &&
            !_tcsncmp(cp, TEXT("Host"), _tcslen(TEXT("Host")))) {
            temp = GetOption(cp, TEXT("Host"));
            if (!temp) {
                continue;
            }
            if (_tcschr(temp, TEXT('*')) || _tcschr(temp, TEXT(' '))) {
                continue;
            }

            if (CB_ERR == SendDlgItemMessage(g_hWnd, IDC_PF_HOSTCOMBO,
                                             CB_FINDSTRINGEXACT, (WPARAM)0,
                                             (LPARAM)temp)) {
                SendDlgItemMessage(g_hWnd, IDC_PF_HOSTCOMBO, CB_ADDSTRING,
                                   (WPARAM)0, (LPARAM)temp);
            }
        }
    }

    fclose(file);
}

/* Initializes the main dialog. */
static BOOL InitDialog()
{
    HICON hIcon;
    TCHAR buf[MAX_PATH];

    if (!LoadInfo()) {
        MessageBox(NULL, TEXT("Failed to get info from ini file."), APP_NAME,
                   MB_OK | MB_ICONERROR);
        Exit();
    }

    SendDlgItemMessage(g_hWnd, IDC_PF_SHOWMESSAGECHK, BM_SETCHECK,
                       (WPARAM)g_showMessages[0], (LPARAM)0);

    if (!_tcscmp(g_configFolder, TEXT(""))) {
        SendDlgItemMessage(g_hWnd, IDC_PF_CONFIGFILEEDIT, WM_SETTEXT,
                           (WPARAM)0, (LPARAM)TEXT(""));
    }
    else {
        _tcscpy(buf, g_configFolder);
        if (g_configFolder[_tcslen(g_configFolder) - 1] != TEXT('\\')) {
            _tcscat(buf, TEXT("\\"));
        }
        _tcscat(buf, g_configFile);
        SendDlgItemMessage(g_hWnd, IDC_PF_CONFIGFILEEDIT, WM_SETTEXT,
                           (WPARAM)0, (LPARAM)buf);
    }

    CreateHostsList(buf);

    SendDlgItemMessage(g_hWnd, IDC_PF_AUTOHIDECHK, BM_SETCHECK,
                       (WPARAM)g_automaticallyHide, (LPARAM)0);

    SendDlgItemMessage(g_hWnd, IDC_PF_STATUSSTATIC, WM_SETTEXT,
                       (WPARAM)0, (LPARAM)PF_STRING_NOT_CONNECTED);

    EnableWindow(GetDlgItem(g_hWnd, IDC_PF_HIDEBTN), FALSE);

    hIcon = LoadImage(g_hInst, MAKEINTRESOURCE(IDI_PF_ICON),
                      IMAGE_ICON, 16, 16, 0);
    PostMessage(g_hWnd, WM_SETICON, FALSE, (LPARAM)hIcon);

    return TRUE;
}

/* Dialog box procedure for input dialog. */
static BOOL CALLBACK InputDlgProc(HWND hWnd,
                                  UINT wMsg,
                                  WPARAM wParam,
                                  LPARAM lParam)
{
    static struct inputinfo *info;

    switch (wMsg) {
    case WM_INITDIALOG:
    {
        SetForegroundWindow(g_hWnd);
        info = (struct inputinfo*)lParam;
        SendMessage(hWnd, WM_SETTEXT, 0, (LPARAM)APP_NAME);
        SendDlgItemMessage(hWnd, IDC_PF_INPUTSTATIC,
                           WM_SETTEXT, 0, (LPARAM)info->string);
        if (info->passwd)
            SendDlgItemMessage(hWnd, IDC_PF_INPUTEDIT, EM_SETPASSWORDCHAR,
                               (WPARAM)TEXT('*'), 0);
        return TRUE;
    }
    case WM_COMMAND:
        switch (LOWORD(wParam)) {
        case IDOK:
        {
            TCHAR inputstring[MAX_INPUT];
            char *inputstring2;
            EnableWindow(GetDlgItem(hWnd, IDOK), FALSE);
            SendDlgItemMessage(hWnd, IDC_PF_INPUTEDIT, WM_GETTEXT,
                               (WPARAM)MAX_INPUT, (LPARAM)inputstring);
            inputstring2 = lpstrCreate(inputstring);
            strcpy(info->inputstring, inputstring2);
            free(inputstring2);
            EndDialog(hWnd, IDOK);
            break;
        }
        default:
            break;
        }
    default:
        break;
    }

    return FALSE;
}

/* Dialog box procedure for input dialog2. */
static BOOL CALLBACK InputDlgProc2(HWND hWnd,
                                   UINT wMsg,
                                   WPARAM wParam,
                                   LPARAM lParam)
{
    switch (wMsg) {
    TCHAR inputstring[MAX_INPUT];
    case WM_INITDIALOG:
    {
#if (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300))
        /* for Pocket PC */
        SHINITDLGINFO shidi;

        /* for Pocket PC */
        /* Create a Done button and size it full screen. */
        shidi.dwMask = SHIDIM_FLAGS;
        shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIZEDLGFULLSCREEN;
        shidi.hDlg = hWnd;
        SHInitDialog(&shidi);
#endif /* (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300)) */
        g_hTtyMsgDlg = hWnd;
        return TRUE;
    }
    case PF_ADDTTYMESSAGE_MSG:
    {
        LPTSTR message;
        DWORD rtn;

        SetForegroundWindow(hWnd);
        message = (LPTSTR)wParam;
        rtn = SendDlgItemMessage(hWnd, IDC_PF_TTYMESSAGEEDIT,
                                 WM_GETTEXTLENGTH, 0, 0);
        SendDlgItemMessage(hWnd, IDC_PF_TTYMESSAGEEDIT, EM_SETSEL, -1, rtn);
        SendDlgItemMessage(hWnd, IDC_PF_TTYMESSAGEEDIT, EM_REPLACESEL, 0,
                           (LPARAM)message);
        if (_tcsstr(message, TEXT("passphrase")) ||
            _tcsstr(message, TEXT("password"))) {
            SendDlgItemMessage(hWnd, IDC_PF_PASSWORDCHK, BM_SETCHECK,
                               (WPARAM)BST_CHECKED, 0);
            SendDlgItemMessage(hWnd, IDC_PF_INPUTEDIT, EM_SETPASSWORDCHAR,
                               (WPARAM)TEXT('*'), 0);
        }
        else {
            SendDlgItemMessage(hWnd, IDC_PF_PASSWORDCHK, BM_SETCHECK,
                               (WPARAM)BST_UNCHECKED, 0);
            SendDlgItemMessage(hWnd, IDC_PF_INPUTEDIT, EM_SETPASSWORDCHAR,
                               0, 0);
        }
        free(message);
        return TRUE;
    }
    case WM_COMMAND:
        switch (LOWORD(wParam)) {
        case IDOK:
        {
            char *inputstring2;
            EnableWindow(GetDlgItem(hWnd, IDOK), FALSE);
            SendDlgItemMessage(hWnd, IDC_PF_INPUTEDIT, WM_GETTEXT,
                               (WPARAM)(MAX_INPUT - 1), (LPARAM)inputstring);
            _tcscat(inputstring, TEXT("\n"));
            inputstring2 = lpstrCreate(inputstring);
            if (strlen(inputstring2) != strlen("\n")) {
                send_string_to_server(inputstring2);
            }
            else {
                if (BST_CHECKED == SendDlgItemMessage(hWnd,
                                                      IDC_PF_PASSWORDCHK,
                                                      BM_GETCHECK, 0, 0)) {
                    /* Probably NULL passphrase. */
                    send_string_to_server(inputstring2);
                }
            }
            free(inputstring2);
            ShowWindow(hWnd, SW_HIDE);
            SendDlgItemMessage(hWnd, IDC_PF_INPUTEDIT, WM_SETTEXT,
                               0, (LPARAM)TEXT(""));
            SendDlgItemMessage(hWnd, IDC_PF_TTYMESSAGEEDIT, WM_SETTEXT,
                               0, (LPARAM)TEXT(""));
            EnableWindow(GetDlgItem(hWnd, IDOK), TRUE);
            break;
        }
        case IDC_PF_PASSWORDCHK:
            if (BST_CHECKED == SendDlgItemMessage(hWnd, IDC_PF_PASSWORDCHK,
                                                  BM_GETCHECK, 0, 0)) {
                SendDlgItemMessage(hWnd, IDC_PF_INPUTEDIT, EM_SETPASSWORDCHAR,
                                   (WPARAM)TEXT('*'), 0);
            }
            else {
                SendDlgItemMessage(hWnd, IDC_PF_INPUTEDIT, EM_SETPASSWORDCHAR,
                                   0, 0);
            }
            break;
        default:
            break;
        }
    default:
        break;
    }

    return FALSE;
}

/* Dialog box procedure. */
static BOOL CALLBACK MainDlgProc(HWND hWnd,
                                 UINT wMsg,
                                 WPARAM wParam,
                                 LPARAM lParam)
{
    switch (wMsg) {
    case WM_INITDIALOG:
    {
#if (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300))
        /* for Pocket PC */
        SHINITDLGINFO shidi;
#endif /* (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300)) */

        g_hWnd = hWnd;

#if (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300))
        /* for Pocket PC */
        /* Create a Done button and size it full screen. */
        shidi.dwMask = SHIDIM_FLAGS;
        shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIZEDLGFULLSCREEN;
        shidi.hDlg = hWnd;
        SHInitDialog(&shidi);
#endif /* (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300)) */

        if (!InitDialog()) {
            MessageBox(NULL, TEXT("Initialization failed!"),
                       APP_NAME, MB_OK | MB_ICONERROR);
            Exit();
        }

#ifdef _LOG_
        OpenLogFile();
#endif /* _LOG_ */
        return TRUE;
    }
    case WM_DESTROY:
        if (g_hTtyMsgDlg) {
            DestroyWindow(g_hTtyMsgDlg);
            g_hTtyMsgDlg = NULL;
        }
        PostQuitMessage(0);
        return TRUE;
    case PF_CONNECTED_MSG:
    {
        DWORD threadID;
        CloseHandle(g_hMainThread);
        g_hMainThread = NULL;
        SendDlgItemMessage(g_hWnd, IDC_PF_STATUSSTATIC, WM_SETTEXT,
                           (WPARAM)0, (LPARAM)PF_STRING_CONNECTED);
        EnableWindow(GetDlgItem(g_hWnd, IDC_PF_HIDEBTN), TRUE);
        g_hClientLoopThread = CreateThread(NULL, 0, DoClientLoop,
                                           NULL, 0, &threadID);
        if (g_automaticallyHide) {
            ShowWindow(g_hWnd, SW_HIDE);
            SetIcon();
        }
        return TRUE;
    }
    case PF_FATALCALLED_MSG:
        Exit();
        return TRUE;
    case PF_TASKBAR_MSG:
        switch (lParam) {
        case WM_LBUTTONDOWN:
        case WM_RBUTTONDOWN:
        {
            HMENU menu;
            HMENU menuTrackPopup;
            POINT curPos;
            static BOOL menuFlag = FALSE;

            EnterCriticalSection(&critical_section);
            if (menuFlag) {
                LeaveCriticalSection(&critical_section);
                return TRUE;
            }
            else {
                menuFlag = TRUE;
                LeaveCriticalSection(&critical_section);
            }

            menu = LoadMenu(g_hInst, MAKEINTRESOURCE(IDR_PF_MENU));
            menuTrackPopup = GetSubMenu(menu, 0);
#ifndef _WIN32_WCE
            GetCursorPos(&curPos);
#else /* _WIN32_WCE */
            curPos.x = GetSystemMetrics(SM_CXSCREEN);
            curPos.y = GetSystemMetrics(SM_CYSCREEN)
                       - GetSystemMetrics(SM_CYCAPTION);
#endif /* _WIN32_WCE */

            SetForegroundWindow(g_hWnd);

            TrackPopupMenuEx(menuTrackPopup, TPM_LEFTALIGN |TPM_BOTTOMALIGN,
                             curPos.x, curPos.y, g_hWnd, NULL);

            PostMessage(g_hWnd, WM_USER, 0, 0);
            DestroyMenu(menu);

            EnterCriticalSection(&critical_section);
            menuFlag = FALSE;
            LeaveCriticalSection(&critical_section);
            return TRUE;
        }
        default:
            break;
        }

        break;
    case PF_GETSTRING_MSG:
        DialogBoxParam(g_hInst, MAKEINTRESOURCE(IDD_PF_INPUTDLG),
                       g_hWnd, InputDlgProc, (LPARAM)wParam);
        return TRUE;
    case PF_SHOWTTYMSGDLG_MSG:
        if (!g_hWnd) {
            return TRUE;
        }
        if (!g_hTtyMsgDlg) {
            g_hTtyMsgDlg = CreateDialog(g_hInst,
                                        MAKEINTRESOURCE(IDD_PF_TTYMSGDLG),
                                        g_hWnd, InputDlgProc2);
            if (!g_hTtyMsgDlg) {
                MessageBox(NULL, TEXT("Could not create \"Show server")
                           TEXT(" message dialog\"."), APP_NAME,
                           MB_OK | MB_ICONERROR);
                Exit();
            }
        }
        ShowWindow(g_hTtyMsgDlg, SW_SHOWNORMAL);
        PostMessage(g_hTtyMsgDlg, PF_ADDTTYMESSAGE_MSG, wParam, 0);
        return TRUE;
    case PF_SHOWMESSAGEBOX_MSG:
    {
        struct msgboxinfo *info;
        info = (struct msgboxinfo*)wParam;
        SetForegroundWindow(g_hWnd);
        info->result =  MessageBox(g_hWnd, info->message,
                                   info->caption, info->type);
        return TRUE;
    }
    case WM_COMMAND:
        switch (LOWORD(wParam)) {
        case IDC_PF_CONNECTBTN:
            if (GetInfo()) {
                EnableWindow(GetDlgItem(g_hWnd, IDC_PF_CONNECTBTN), FALSE);
                EnableWindow(GetDlgItem(g_hWnd, IDC_PF_HOSTCOMBO), FALSE);
                EnableWindow(GetDlgItem(g_hWnd, IDC_PF_SHOWMESSAGECHK), FALSE);
                EnableWindow(GetDlgItem(g_hWnd, IDC_PF_CONFIGFILEBTN), FALSE);
                EnableWindow(GetDlgItem(g_hWnd, IDC_PF_AUTOHIDECHK), FALSE);
                SaveInfo();
                AdjustFileNames();
                Connect();
                return TRUE;
            }
            else return FALSE;
#if (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300))
        case IDOK:
            /* for Pocket PC */
#endif /* (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300)) */
        case IDCANCEL:
            EnableWindow(GetDlgItem(g_hWnd, IDCANCEL), FALSE);
#if (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300))
            EnableWindow(GetDlgItem(g_hWnd, IDOK), FALSE);
#endif /* (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300)) */
            if (connectionStatus == PF_STATUS_NOT_CONNECTED) {
                DestroyWindow(g_hWnd);
            }
            else if (connectionStatus == PF_STATUS_CONNECTING) {
                MessageBox(g_hWnd, TEXT("Connection canceled by user."),
                           APP_NAME, MB_OK | MB_ICONEXCLAMATION);
                CloseHandle(g_hMainThread);
#ifdef _WIN32_WCE
                ExitThread(0);
#else /* _WIN32_WCE */
                ExitProcess(0);
#endif /* _WIN32_WCE */
            }
            else if (connectionStatus == PF_STATUS_CONNECTED) {
                LPTSTR buf;
                int rtn;

                buf = FmtMsg(TEXT("Connection : %1!s!\n\n")
                             TEXT("Do you really want to disconnect?"),
                             hostTemp);
                rtn = MessageBox(g_hWnd, buf, APP_NAME,
                                 MB_YESNO | MB_ICONQUESTION);
                LocalFree(buf);
                if (rtn == IDYES) {
                    if (g_hClientLoopThread) {
                        close_now();
                        PrepareExit();
                    }
#ifdef _LOG_
                    CloseLogFile();
#endif /* _LOG_ */
                    RemoveIcon();
                    DestroyWindow(g_hWnd);
                }
                else {
                    EnableWindow(GetDlgItem(g_hWnd, IDCANCEL), TRUE);
#if (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300))
                    EnableWindow(GetDlgItem(g_hWnd, IDOK), TRUE);
#endif /* (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300)) */
                }
            }
            return TRUE;
        case IDC_PF_CONFIGFILEBTN:
        {
            TCHAR fileBuf[MAX_PATH];
            fileBuf[0] = TEXT('\0');
            if (FindFile(g_hWnd, TEXT("Open: Config File"),
                         TEXT("All Files\0*.*\0\0"),
                         g_configFolder, fileBuf, MAX_PATH)) {
                LPTSTR temp;
                _tcscpy(g_configFolder, fileBuf);
                SendDlgItemMessage(g_hWnd, IDC_PF_CONFIGFILEEDIT, WM_SETTEXT,
                                   (WPARAM)0, (LPARAM)g_configFolder);

                CreateHostsList(g_configFolder);

                temp = _tcsrchr(g_configFolder, TEXT('\\'));
                *temp = TEXT('\0');
                _tcscpy(g_configFile, temp + 1);
                temp = _tcschr(g_configFolder, TEXT('\\'));
                if (!temp) {
                    /* Must be root directory. */
                    _tcscat(g_configFolder, TEXT("\\"));
                }
            }
            return TRUE;
        }
        case IDC_PF_HOSTCOMBO:
        {
            int i;
            if ((HIWORD(wParam) == CBN_EDITCHANGE)) {
                SendDlgItemMessage(g_hWnd, IDC_PF_HOSTCOMBO, WM_GETTEXT,
                                   (WPARAM)NI_MAXHOST, (LPARAM)hostTemp);
            }
            else if (HIWORD(wParam) == CBN_SELCHANGE) {
                int rtn;
                rtn = SendDlgItemMessage(g_hWnd, IDC_PF_HOSTCOMBO,
                                         CB_GETCURSEL, 0, 0);

                SendDlgItemMessage(g_hWnd, IDC_PF_HOSTCOMBO, CB_GETLBTEXT,
                                   (WPARAM)rtn, (LPARAM)hostTemp);
            }
            else {
                return TRUE;
            }

            SendDlgItemMessage(g_hWnd, IDC_PF_SHOWMESSAGECHK,
                               BM_SETCHECK, (WPARAM)FALSE,
                               (LPARAM)0);
            i = 0;
            while (_tcscmp(g_hosts[i], TEXT(""))) {
                if (!_tcscmp(g_hosts[i], hostTemp) && g_showMessages[i]) {
                    SendDlgItemMessage(g_hWnd, IDC_PF_SHOWMESSAGECHK,
                                       BM_SETCHECK, (WPARAM)TRUE,
                                       (LPARAM)0);
                    break;
                }
                i ++;
            }
            return TRUE;
        }
        case IDC_PF_HIDEBTN:
            ShowWindow(g_hWnd, SW_HIDE);
            SetIcon();
            return TRUE;
        case ID_PF_SHOW:
            RemoveIcon();
            ShowWindow(g_hWnd, SW_SHOW);
            SetForegroundWindow(g_hWnd);
            return TRUE;
        default:
            break;
        }

        break;
    default:
        break;
    }

    return FALSE;
}

/* Main function of the program. */
int WINAPI WinMain(HINSTANCE hInstance,
                   HINSTANCE hPrevInstance,
                   LPTSTR lpCmdLine,
                   int nCmdShow)
{
    static WSADATA wsaData;
    BOOL quit = FALSE;
    MSG msg;

    if (hPrevInstance)
        return FALSE;

    g_hInst = hInstance;
    host = NULL;
    InitializeCriticalSection(&critical_section);

    wsaStartup = WSAStartup(MAKEWORD(1, 1), &wsaData);
    if (wsaStartup) {
        MessageBox(NULL, TEXT("Could not initialize winsock."),
                   APP_NAME, MB_OK | MB_ICONERROR);
        Exit();
    }

    g_hWnd = CreateDialog(g_hInst, MAKEINTRESOURCE(IDD_PF_MAINDLG),
                          NULL, MainDlgProc);
    ShowWindow(g_hWnd, SW_SHOWNORMAL);

    while (GetMessage(&msg, NULL, 0, 0)) {
        if ((g_hWnd == NULL || !IsDialogMessage(g_hWnd, &msg)) &&
            (g_hTtyMsgDlg == NULL ||
             !IsDialogMessage(g_hTtyMsgDlg, &msg))) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }

    if (!wsaStartup) {
        WSACleanup();
    }

    DeleteCriticalSection(&critical_section);

    if (host) {
        free(host);
    }

    return FALSE;
}
