/*
 * Copyright (C) 2000 Fujio Nobori (toh@fuji-climb.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pf-keygen.c,v 1.3 2002/02/06 23:28:45 toh Exp $
 */

#include <windows.h>
#include "rsa.h"
#include "PortForwarder.h"
#include "resource2.h"
#include "ssh.h"
#include "xmalloc.h"
#if (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300))
#include <aygshell.h>
#endif /* (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300)) */

#define INFO_SIZE 1024
/* Handle to current instance. */
HINSTANCE g_hInst;
/* Handle to main dialog. */
HWND g_hWnd;
/* Flag indicating whether user requests to show tty messages. */
BOOL g_showMessage;
/* Hosts that were most currently used. */
TCHAR g_hosts[MAX_HOST_NUM][MAX_PATH];
/* Statuses of show message that were most currently used. */
BOOL g_showMessages[MAX_HOST_NUM];
/* Path of config folder. */
TCHAR g_configFolder[MAX_PATH];
/* Name of knowon-hosts file. */
TCHAR g_knownHostsFile[MAX_PATH];
/* Name of identity file. */
TCHAR g_identityFile[MAX_PATH];
/* Name of config file. */
TCHAR g_configFile[MAX_PATH];
/* Flag indicating whether user requests to hide main dialog automatically. */
BOOL g_automaticallyHide;
/* Name of the host we are connecting to. */
char *host;

/* The identity file name, given on the command line or entered by the user. */
extern char identity_file[1024];

/* Struct for InputDlg */
struct InputDlgInfo {
    TCHAR caption[INFO_SIZE];
    TCHAR text[INFO_SIZE];
    TCHAR edit[INFO_SIZE];
    BOOL isPassphrase;
};

static struct InputDlgInfo g_dlgInfo;

void do_fingerprint();
void do_change_comment();
void do_change_passphrase();
void do_keygen();

/* Dialog box procedure. */
static BOOL CALLBACK InputDlgProc(HWND hWnd,
                                  UINT wMsg,
                                  WPARAM wParam,
                                  LPARAM lParam)
{
    switch (wMsg) {
    case WM_INITDIALOG:
    {
        SendMessage(hWnd, WM_SETTEXT, 0, (LPARAM)g_dlgInfo.caption);
        SendDlgItemMessage(hWnd, IDC_PFKG_STATIC, WM_SETTEXT, 0,
                           (LPARAM)g_dlgInfo.text);
        if (g_dlgInfo.isPassphrase) {
            SendDlgItemMessage(hWnd, IDC_PFKG_EDIT, EM_SETPASSWORDCHAR,
                               (WPARAM)'*', 0);
        }
        return TRUE;
    }
    case WM_COMMAND:
        switch (LOWORD(wParam)) {
            case IDOK:
                SendDlgItemMessage(hWnd, IDC_PFKG_EDIT, WM_GETTEXT,
                                   (WPARAM)INFO_SIZE, (LPARAM)g_dlgInfo.edit);
                EndDialog(hWnd, IDOK);
                break;
            case IDCANCEL:
                EndDialog(hWnd, IDCANCEL);
                break;
            default:
                break;
        }
    default:
        break;
    }

    return FALSE;
}

/* Get new comment */
BOOL GetComment(LPCTSTR caption,
                LPCTSTR text,
                char *comment)
{
    int rtn;
    _tcscpy(g_dlgInfo.caption, caption);
    _tcscpy(g_dlgInfo.text, text);
    g_dlgInfo.isPassphrase = FALSE;

    rtn = DialogBox(g_hInst, MAKEINTRESOURCE(IDD_PFKG_INPUTDLG),
                    g_hWnd, InputDlgProc);
    if (rtn == IDCANCEL) {
        MessageBox(g_hWnd, TEXT("Canceled by user."),
                   APP_NAME2, MB_OK | MB_ICONERROR);
        return FALSE;
    }

    lpstrCopy(comment, 1024, g_dlgInfo.edit);
    return TRUE;
}

/* Get passphrase */
char *GetPassphrase(LPCTSTR caption,
                    LPCTSTR text)
{
    char *rtnString;
    int rtn;
    _tcscpy(g_dlgInfo.caption, caption);
    _tcscpy(g_dlgInfo.text, text);
    g_dlgInfo.isPassphrase = TRUE;

    rtn = DialogBox(g_hInst, MAKEINTRESOURCE(IDD_PFKG_INPUTDLG),
                    g_hWnd, InputDlgProc);
    if (rtn == IDCANCEL) {
        MessageBox(g_hWnd, TEXT("Canceled by user."),
                   APP_NAME2, MB_OK | MB_ICONERROR);
        return NULL;
    }

    rtnString = xmalloc(_tcslen(g_dlgInfo.edit) + 1);
    lpstrCopy(rtnString, _tcslen(g_dlgInfo.edit) + 1, g_dlgInfo.edit);
    return rtnString;
}

/* Dialog box procedure. */
static BOOL CALLBACK MainDlgProc(HWND hWnd,
                                 UINT wMsg,
                                 WPARAM wParam,
                                 LPARAM lParam)
{
    switch (wMsg) {
    case WM_INITDIALOG:
    {
        HICON hIcon;
#if (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300))
        /* for Pocket PC */
        SHINITDLGINFO shidi;
#endif /* (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300)) */

        g_hWnd = hWnd;

#if (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300))
        /* for Pocket PC */
        /* Create a Done button and size it full screen. */
        shidi.dwMask = SHIDIM_FLAGS;
        shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIZEDLGFULLSCREEN;
        shidi.hDlg = hWnd;
        SHInitDialog(&shidi);
#endif /* (defined(_WIN32_WCE_PSPC) && (_WIN32_WCE >= 300)) */

        SendDlgItemMessage(hWnd, IDC_PFKG_GENKEYRADIO, BM_SETCHECK,
                           (WPARAM)BST_CHECKED, 0);
        {
            LoadInfo();
            if (_tcscmp(g_configFolder, TEXT(""))) {
                TCHAR buf[MAX_PATH];
                _tcscpy(buf, g_configFolder);
                if (g_configFolder[_tcslen(g_configFolder) - 1] != TEXT('\\')) {
                    _tcscat(buf, TEXT("\\"));
                }
                _tcscat(buf, g_identityFile);
                SendDlgItemMessage(hWnd, IDC_PFKG_FILEEDIT, WM_SETTEXT,
                                   0, (LPARAM)buf);
            }
        }
        hIcon = LoadImage(g_hInst, MAKEINTRESOURCE(IDI_PFKG_ICON),
                          IMAGE_ICON, 16, 16, 0);
        PostMessage(hWnd, WM_SETICON, FALSE, (LPARAM)hIcon);
        return TRUE;
    }
    case WM_COMMAND:
        switch (LOWORD(wParam)) {
        case IDC_PFKG_FILEBUTTON:
        {
            TCHAR fileBuf[MAX_PATH];
            fileBuf[0] = TEXT('\0');
            if (FindFile(g_hWnd, TEXT("Open: Config File"),
                         TEXT("All Files\0*.*\0\0"),
                         g_configFolder, fileBuf, MAX_PATH)) {
                LPTSTR temp;
                _tcscpy(g_configFolder, fileBuf);
                SendDlgItemMessage(g_hWnd, IDC_PFKG_FILEEDIT, WM_SETTEXT,
                                   (WPARAM)0, (LPARAM)g_configFolder);

                temp = _tcsrchr(g_configFolder, TEXT('\\'));
                *temp = TEXT('\0');
                _tcscpy(g_identityFile, temp + 1);
                temp = _tcschr(g_configFolder, TEXT('\\'));
                if (!temp) {
                    /* Must be root directory. */
                    _tcscat(g_configFolder, TEXT("\\"));
                }
            }
            return TRUE;
        }
        case IDOK:
            SendDlgItemMessage(hWnd, IDC_PFKG_FILEEDIT, WM_GETTEXT,
                               (WPARAM)MAX_PATH, (LPARAM)g_identityFile);
            lpstrCopy(identity_file, 1024, g_identityFile);

            if (!_tcscmp(g_identityFile, TEXT(""))) {
                MessageBox(hWnd, TEXT("Please specify \"Identity file\"."),
                           APP_NAME2, MB_OK | MB_ICONEXCLAMATION);
                EnableWindow(GetDlgItem(hWnd, IDOK), TRUE);
            }
            else if (BST_CHECKED == SendDlgItemMessage(hWnd,
                                                       IDC_PFKG_GENKEYRADIO,
                                                       BM_GETSTATE, 0, 0)) {
                do_keygen();
            }
            else if (BST_CHECKED == SendDlgItemMessage(hWnd,
                                                       IDC_PFKG_CHGPASSRADIO,
                                                       BM_GETSTATE, 0, 0)) {
                do_change_passphrase();
            }
            else if (BST_CHECKED == SendDlgItemMessage(hWnd,
                                                       IDC_PFKG_CHGCOMMENTRADIO,
                                                       BM_GETSTATE, 0, 0)) {
                do_change_comment();
            }
            else if (BST_CHECKED == SendDlgItemMessage(hWnd,
                                                       IDC_PFKG_SHOWFINGERRADIO,
                                                       BM_GETSTATE, 0, 0)) {
                do_fingerprint();
            }

            return TRUE;
        case IDCANCEL:
            EndDialog(hWnd, 0);
            return TRUE;
#if 0
        case IDC_PF_CONFIGFILEBTN:
        {
            TCHAR fileBuf[MAX_PATH];
            fileBuf[0] = TEXT('\0');
            if (FindFile(g_hWnd, TEXT("Open: Config File"),
                         TEXT("All Files\0*.*\0\0"),
                         g_configFolder, fileBuf, MAX_PATH)) {
                LPTSTR temp;
                _tcscpy(g_configFolder, fileBuf);
                SendDlgItemMessage(g_hWnd, IDC_PF_CONFIGFILEEDIT, WM_SETTEXT,
                                   (WPARAM)0, (LPARAM)g_configFolder);

                CreateHostsList(g_configFolder);

                temp = _tcsrchr(g_configFolder, TEXT('\\'));
                *temp = TEXT('\0');
                _tcscpy(g_configFile, temp + 1);
                temp = _tcschr(g_configFolder, TEXT('\\'));
                if (!temp) {
                    /* Must be root directory. */
                    _tcscat(g_configFolder, TEXT("\\"));
                }
            }
            return TRUE;
        }
        case IDC_PF_HOSTCOMBO:
        {
            int i;
            if ((HIWORD(wParam) == CBN_EDITCHANGE)) {
                SendDlgItemMessage(g_hWnd, IDC_PF_HOSTCOMBO, WM_GETTEXT,
                                   (WPARAM)NI_MAXHOST, (LPARAM)hostTemp);
            }
            else if (HIWORD(wParam) == CBN_SELCHANGE) {
                int rtn;
                rtn = SendDlgItemMessage(g_hWnd, IDC_PF_HOSTCOMBO,
                                         CB_GETCURSEL, 0, 0);

                SendDlgItemMessage(g_hWnd, IDC_PF_HOSTCOMBO, CB_GETLBTEXT,
                                   (WPARAM)rtn, (LPARAM)hostTemp);
            }
            else {
                return TRUE;
            }

            SendDlgItemMessage(g_hWnd, IDC_PF_SHOWMESSAGECHK,
                               BM_SETCHECK, (WPARAM)FALSE,
                               (LPARAM)0);
            i = 0;
            while (_tcscmp(g_hosts[i], TEXT(""))) {
                if (!_tcscmp(g_hosts[i], hostTemp) && g_showMessages[i]) {
                    SendDlgItemMessage(g_hWnd, IDC_PF_SHOWMESSAGECHK,
                                       BM_SETCHECK, (WPARAM)TRUE,
                                       (LPARAM)0);
                    break;
                }
                i ++;
            }
            return TRUE;
        }
        case IDC_PF_HIDEBTN:
            ShowWindow(g_hWnd, SW_HIDE);
            SetIcon();
            return TRUE;
        case ID_PF_SHOW:
            RemoveIcon();
            ShowWindow(g_hWnd, SW_SHOW);
            SetForegroundWindow(g_hWnd);
            return TRUE;
#endif /* 0 */
        default:
            break;
        }

        break;
    default:
        break;
    }

    return FALSE;
}

/* Main function of the program. */
int WINAPI WinMain(HINSTANCE hInstance,
                   HINSTANCE hPrevInstance,
                   LPTSTR lpCmdLine,
                   int nCmdShow)
{
    if (hPrevInstance)
        return FALSE;

    g_hInst = hInstance;

    /* check if RSA support exists */
    if (rsa_alive() == 0) {
        MessageBox(NULL, TEXT("rsa_alive() failed."), APP_NAME2,
                   MB_OK | MB_ICONERROR);
        return FALSE;
    }

    DialogBox(g_hInst, MAKEINTRESOURCE(IDD_PFKG_MAINDLG), NULL, MainDlgProc);

    return FALSE;
}
